/*
 * MAX : zfk_stepper.cc, Wed Jan  2 14:44:59 CET 2002 -sg
 *
 * Device driver for the ZFK Steppercontroller
 * 	only one stepper motor on one ZFK-module
 *
 * M. Dammasch	08.09.1998 upgrade to max2.0beta
 * M. Dammasch   24.11.1998 error()-handling changed (only 0 returns)
 * M. Dammasch   03.06.1999 define unique DEBUG-Flag for verbose/silent mode
 * M. Dammasch   01.06.2000 add "shots" variable
 * 
 * This file is part of Max data acquisition software
 * (c) 1997-98 by D. Rohloff & M. Dammasch
 *
 * Max is free software; you can redistribute it and/or modify it
 * under the terms of the version 2 of GNU General Public License as
 * published by the Free Software Foundation.
 *
 * Max is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * (see the file COPYING) in this directory; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>
#include "list.hh"
#include "sched.hh"
#include "camacs.h"
#include "error.hh"


#define DEBUG_DEV_ZFK

/*
    stepping frequencies [Hz]
    (different for each CAMAC-Modul - hardware set!)
*/
#define FREQ_START  100  // static mode: const.: FREQ_START
#define FREQ_MAX    300  // dynamic mode: accel.: FREQ_START->FREQ_MAX

#define WAIT        100   // delay between stepper commands
                          //(due to timing problems by the controller

/*
    Moving direction
*/
#define LEFT   (0x1 << 23)  // Bit W24 selected
#define RIGHT   0x0	    // Bit W24 not selected

/*
    Stepping mode
*/
#define WITH_DYNAMIC_CONTROLL 	 (0x1 << 22)  // Bit W23 selected
#define WITHOUT_DYNAMIC_CONTROLL  0x0         // Bit W23 not selected

/*
    Step motor driver bits, i.e address
*/
#define DRIVER_ADDR_3  3145728  // Bits W22 And W21 selected
#define DRIVER_ADDR_2  2097152  // Bit  W22 selected (only)
#define DRIVER_ADDR_1  1048576  // Bit  W21 selected (only)
#define DRIVER_ADDR_0  0        // Bits W22 and W21 not selected

/*
	LAM-Mask Register bits
	Read: F(1)A(13), Write: F(17)A(13)
*/
#define LAM_AT_STEP_REG_EMPTY   0x1        // W1 selected
#define LAM_AT_RIGHT_END_POS   (0x1 << 1)  // W2 selected
#define LAM_AT_LEFT_END_POS    (0x1 << 2)  // W3 selected
#define LAM_AT_STOP_INPUT      (0x1 << 3)  // W4 selected

/*
    Global variable(s)
*/
static int TLAM;		// timeout (general value- device 'all')
static int N;			// Crate Slot No.
static int drv_addr;            // address of physical step motor driver
static int mode;                // stepping mode
static int single_step;         // flag for single step mode (mode 0)
static int forward, backward;	// moving direction
static int steps_inc      = 0;  // step size (per channel)
static int position	  = 0;  // new position
static int old_position   = 0;  // actual position
static int start_position = 0;  // position at startup
static int shots;		// move only after "shots" channels

int zfk_stepper_init(List *device_data, Scheduler *parameters)
{
  unsigned int dummy;
  int qx=0, direction, dd;
	
  if (camac_init() != 1)
    return parameters->error(E_CAMAC,"Couldn't initialize device ");

  printf("\n\033[0minitializing zfk_stepper ");	

  device_data->getvar(&TLAM,"all","timeout");  // get general timeout value
  //printf("\nTLAM = %d ",TLAM);

  device_data->getvar(&N,"zfk_stepper","N");  // get number of CAMAC station
  //printf("\nN = %d ",N);

  device_data->getvar(&dd,"zfk_stepper","Driver");  // get address of physical step motor driver
  switch (dd) {
  case 1: drv_addr = DRIVER_ADDR_0; break;
  case 2: drv_addr = DRIVER_ADDR_1; break;
  case 3: drv_addr = DRIVER_ADDR_2; break;
  case 4: drv_addr = DRIVER_ADDR_3; break;
  default: return parameters->error(E_ALLFUCKEDUP,"zfk_stepper: Illegal driver address"); break;
  }
  //printf("\nDriver = %d (0x%x) ",dd,drv_addr);

  device_data->getvar(&dd,"zfk_stepper","Mode");  // get stepping mode
  switch (dd) {
  case 2:
    mode = WITH_DYNAMIC_CONTROLL;
    single_step = 0;
    break;
  case 1:
    mode = WITHOUT_DYNAMIC_CONTROLL;
    single_step = 0;
    break;
  case 0:
    mode = WITHOUT_DYNAMIC_CONTROLL;
    single_step = 1;
    break;
  default: return parameters->error(E_ALLFUCKEDUP,"zfk_stepper: Illegal stepping mode"); break;
  }
  //printf("\nMode = %d (0x%x) ",dd,mode);

  device_data->getvar(&direction,"zfk_stepper","Direction");  // get moving direction (0=right, 1=left)
  //printf("\nDirection = %d ",direction);

  if (direction == 0)
  {
    forward = RIGHT;
    backward = LEFT;
  }
  else
  {
    forward = LEFT;
    backward = RIGHT;
  }

  
  device_data->getvar(&shots,"zfk_stepper","Shots");
  //printf("\nShots = %d ",shots);
  
  device_data->getvar(&steps_inc,"zfk_stepper","Increments");  // get moving steps per increment (per channel)
  //printf("\nIncrements = %d ",steps_inc);

  device_data->getvar(&start_position,"zfk_stepper","Start_pos");  // get position of beginning
  //printf("\nStart_pos = %d ",start_position);

  device_data->getvar(&old_position,"zfk_stepper","Actual_pos");  // get position of beginning
  //printf("\nActual_pos = %d ",old_position);

  dummy = drv_addr;
  //printf("\nzfk_stepper: F(16)A(0)[%d] ",dummy);
  if ((qx = camac(16,0,N,&dummy)) != 3) {  // set driver address
    printf("\nzfk_stepper: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
      return 0;
  }
  //printf("-> QX: %d\tD: %d ",qx,dummy);

  dummy = LAM_AT_STOP_INPUT|LAM_AT_RIGHT_END_POS|LAM_AT_LEFT_END_POS|LAM_AT_STEP_REG_EMPTY;
#ifdef DEBUG_DEV_ZFK
  printf("\nzfk_stepper: F(23)A(12)[%d] ",dummy);
#endif
  if ((qx = camac(23,12,N,&dummy)) != 3) {  // clear LAM-Status register
    printf("\nzfk_stepper: F(23)A(12)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
      return 0;
  }
#ifdef DEBUG_DEV_ZFK
  printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

  /*
  usleep(WAIT);  // timing problem by the controller
  dummy = LAM_AT_STEP_REG_EMPTY;
  //printf("\nzfk_stepper: F(17)A(13)[%d] ",dummy);
  if ((qx = camac(17,13,N,&dummy)) != 3)  // set LAM-Mask register
  {
    printf("\nzfk_stepper: F(17)A(13)[%d] -> QX: %d ",dummy,qx);
    if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
    return 0;
  }
  //printf("-> QX: %d\tD: %d ",qx,dummy);
  */

  printf("-> Pos: %d ",old_position);

  return 1;
} // of init


int zfk_stepper_close(List *device_data, Scheduler *parameters)
{
  printf("\nclosing zfk_stepper -> Pos: %d ",position);

  device_data->putvar(position,"zfk_stepper","Actual_pos");

  if (camac_close() != 1)
    return parameters->error(E_CAMAC,"Couldn't close device ");

  return 1;
}



//device function for zfk_stepper
int zfk_stepper_pre(List *device_data, Scheduler *parameters)
{
  int qx=0, i=0;
  int steps, si, direction;
  unsigned int dummy;

  /* ************ calculate new position  ************* */

  position = (parameters->act_channel()-1) * steps_inc + start_position;

  device_data->getvar(&old_position,"zfk_stepper","Actual_pos");  // get actual position

  steps = position - old_position;

  if (steps >= 0) {
    direction = forward;
  }
  else {
    direction = backward;
  }

  /* ************ moving to new position ************** */

  if ( ! (parameters->act_channel() % shots == 0 && parameters->act_integrations() == 0)) return 1;
  
  if (steps == 0) {  // nothing to do -> quit...
    printf("\nzfk_stepper: ready -> Pos: %d ",position);
    return 1;
  }

  if (single_step) {  // single step mode
    for (si = 0; si < abs(steps); si++) {
      usleep(WAIT);  // timing problem by the controller
      dummy = LAM_AT_STEP_REG_EMPTY;
#ifdef DEBUG_DEV_ZFK
      printf("\nzfk_stepper: F(17)A(13)[%d] ",dummy);
#endif
      if ((qx = camac(17,13,N,&dummy)) != 3) {  // set LAM-Mask register
	printf("\nzfk_stepper: F(17)A(13)[%d] -> QX: %d ",dummy,qx);
	if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
	  return 0;
      }
#ifdef DEBUG_DEV_ZFK
      printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

      dummy = direction + mode + drv_addr + 1;
#ifdef DEBUG_DEV_ZFK
      printf("\n\033[7mStep-Reg: %d (0x%x) + %d (0x%x) + %d (0x%x) + 1 = %d (single step mode)\033[27m",direction,direction,mode,mode,drv_addr,drv_addr,dummy);
      printf("\nzfk_stepper: F(16)A(0)[%d] ",dummy);
#endif
      usleep(WAIT);  // timing problem by the controller
      if ((qx = camac(16,0,N,&dummy)) != 3) {  // set step register
	printf("\nzfk_stepper: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
	if (! (qx&2)) { // q not set - no response
	  if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
	    return 0;
	}
	else {  // qx==0
	  if (!parameters->error(E_ALLFUCKEDUP,"zfk_stepper: F(16)A(0)"))
	    return 0;
	}
      }
#ifdef DEBUG_DEV_ZFK
      printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

      /* now the stepper is moving... */
      printf("\nzfk_stepper: %s%d (%3s) -[%d/%d]-> Pos: %d ",(steps>0)?"+":"-",1,(direction==forward)?"fwd":"rev",si,abs(steps),position);

      // wait until step motor has finished (approx.)
      usleep(WAIT);

      i = 0;
      dummy = 0;
#ifdef DEBUG_DEV_ZFK
      printf("\nzfk_stepper: F(8)A(15)[%d] ",dummy);
#endif
      qx = camac(8,15,N,&dummy);  // test for LAM pending
      while ((++i < TLAM) && (!(qx & 2))) {
	usleep(200);  // timing problem by the controller
	qx = camac(8,15,N,&dummy);
      }
#ifdef DEBUG_DEV_ZFK
      printf("-[%d/%d]-> QX: %d\tD: %d ",i,TLAM,qx,dummy);
#endif
      if (i >= TLAM) {
	if (!parameters->error(E_TIMEOUT,"zfk_stepper: F(8)A(15)"))
	  return 0;
      }

      usleep(WAIT);  // timing problem by the controller
      dummy= 0;
#ifdef DEBUG_DEV_ZFK
      printf("\nzfk_stepper: F(1)A(14)[%d] ",dummy);
#endif
      if ((qx = camac(1,14,N,&dummy)) != 3) {  // read LAM-Request register
	printf("\nzfk_stepper: F(1)A(14)[%d] -> QX: %d ",dummy,qx);
	if (!parameters->error(E_ALLFUCKEDUP,"zfk_stepper: F(1)A(14)"))
	  return 0;
      }
#ifdef DEBUG_DEV_ZFK
      printf("-> QX: %d\tD: %d <%1d%1d%1d%1d> ",qx,dummy,dummy&LAM_AT_STOP_INPUT?1:0,dummy&LAM_AT_RIGHT_END_POS?1:0,dummy&LAM_AT_LEFT_END_POS?1:0,dummy&LAM_AT_STEP_REG_EMPTY?1:0);
#endif

      usleep(WAIT);  // timing problem by the controller
      dummy = LAM_AT_STEP_REG_EMPTY;
#ifdef DEBUG_DEV_ZFK
      printf("\nzfk_stepper: F(23)A(12)[%d] ",dummy);
#endif
      if ( !(qx = camac(23,12,N,&dummy)) & 2) {  // clear LAM-Status register
	printf("\nzfk_stepper: F(23)A(12)[%d] -> QX: %d ",dummy,qx);
	if (!parameters->error(E_ALLFUCKEDUP,"zfk_stepper: F(23)A(12)"))
	  return 0;
      }
#ifdef DEBUG_DEV_ZFK
      printf("-> QX: %d\tD: %d ",qx,dummy);
#endif
    }
  }
  else {
    usleep(WAIT);  // timing problem by the controller
    dummy = LAM_AT_STEP_REG_EMPTY;
#ifdef DEBUG_DEV_ZFK
    printf("\nzfk_stepper: F(17)A(13)[%d] ",dummy);
#endif
    if ((qx = camac(17,13,N,&dummy)) != 3) {  // set LAM-Mask register
      printf("\nzfk_stepper: F(17)A(13)[%d] -> QX: %d ",dummy,qx);
      if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
	return 0;
    }
#ifdef DEBUG_DEV_ZFK
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

    dummy = direction + mode + drv_addr + abs(steps);
#ifdef DEBUG_DEV_ZFK
    printf("\n\033[7mStep-Reg: %d (0x%x) + %d (0x%x) + %d (0x%x) + %d = %d\033[27m ",direction,direction,mode,mode,drv_addr,drv_addr,abs(steps),dummy);
    printf("\nzfk_stepper: F(16)A(0)[%d] ",dummy);
#endif
    usleep(250);  // timing problem by the controller
    if ((qx = camac(16,0,N,&dummy)) != 3) {  // set step register
      printf("\nzfk_stepper: F(16)A(0)[%d] -> QX: %d ",dummy,qx);
      if ( !(qx&2)) { // q not set - no response
	if (!parameters->error(E_DEVICE_NOT_FOUND,"zfk_stepper: No stepper driver detected"))
	  return 0;
      }
      else {  // qx==0
	if (!parameters->error(E_ALLFUCKEDUP,"zfk_stepper: F(16)A(0)"))
	  return 0;
      }
    }
#ifdef DEBUG_DEV_ZFK
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif

    /* now the stepper is moving... */
    printf("\nzfk_stepper: %s%d (%3s) -> Pos: %d ",(steps>0)?"+":"",steps,(direction==forward)?"fwd":"rev",position);

    // wait until step motor has finished (approx.)
    usleep((unsigned long)(1.75E+06*fabs(steps)/FREQ_START));

    i = 0;
    dummy = 0;
#ifdef DEBUG_DEV_ZFK
    printf("\nzfk_stepper: F(8)A(15)[%d] ",dummy);
#endif
    qx = camac(8,15,N,&dummy);  // test for LAM pending
    while ((++i < TLAM) && (!(qx & 2))) {
      usleep(200);  // timing problem by the controller
      qx = camac(8,15,N,&dummy);
    }
#ifdef DEBUG_DEV_ZFK
    printf("-[%d/%d]-> QX: %d\tD: %d ",i,TLAM,qx,dummy);
#endif
    if (i >= TLAM) {
      if (!parameters->error(E_TIMEOUT,"zfk_stepper: F(8)A(15)"))
	return 0;
    }

    usleep(WAIT);  // timing problem by the controller
    dummy= 0;
#ifdef DEBUG_DEV_ZFK
    printf("\nzfk_stepper: F(1)A(14)[%d] ",dummy);
#endif
    if ((qx = camac(1,14,N,&dummy)) != 3) {  // read LAM-Request register
      printf("\nzfk_stepper: F(1)A(14)[%d] -> QX: %d ",dummy,qx);
      if (!parameters->error(E_ALLFUCKEDUP,"zfk_stepper: F(1)A(14)"))
	return 0;
    }	
#ifdef DEBUG_DEV_ZFK
    printf("-> QX: %d\tD: %d <%1d%1d%1d%1d> ",qx,dummy,dummy&LAM_AT_STOP_INPUT?1:0,dummy&LAM_AT_RIGHT_END_POS?1:0,dummy&LAM_AT_LEFT_END_POS?1:0,dummy&LAM_AT_STEP_REG_EMPTY?1:0);
#endif

    usleep(WAIT);  // timing problem by the controller
    dummy = LAM_AT_STEP_REG_EMPTY;
#ifdef DEBUG_DEV_ZFK
    printf("\nzfk_stepper: F(23)A(12)[%d] ",dummy);
#endif
    if ( !(qx = camac(23,12,N,&dummy)) & 2) {  // clear LAM-Status register
      printf("\nzfk_stepper: F(23)A(12)[%d] -> QX: %d ",dummy,qx);
      if (!parameters->error(E_ALLFUCKEDUP,"zfk_stepper: F(23)A(12)"))
	return 0;
    }
#ifdef DEBUG_DEV_ZFK
    printf("-> QX: %d\tD: %d ",qx,dummy);
#endif
  }

  return 1;	
} /* end of zfk_stepper_pre */

int zfk_stepper (List *device_data, Scheduler *parameters)
{
  device_data->putvar(position,"zfk_stepper","Actual_pos");
  return 1;
}











