/*
 * MAX : sched.hh, Sam Dez 29 20:07:54 CET 2001 -sg
 *
 * The scheduler class calls all devices according to their priority list.
 * It also contains the functions for handling the measured data.
 * See the function int Scheduler::store(int )
 *
 * This file is part of Max data acquisition software
 * Copyright (C) 1998 Christian Rosen
 *
 * Max is free software; you can redistribute it and/or modify it
 * under the terms of the version 2 of GNU General Public License as
 * published by the Free Software Foundation.
 *
 * Max is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * (see the file COPYING) in this directory; if not, write to the
 * Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef SCHEDULER_H
#define SCHEDULER_H

#define BACKUP_ARRAY_SIZE MAX_ARRAY_SIZE //size for array to be plotted, rest is truncated
	   			//check with MAX_ARRAY_PLOTSIZE in plot.hpp
#include <qwidget.h>
#include <qtimer.h>
#include <qimage.h>
#include <assert.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

#include "list_polld.h"
#include "error.h"
#include "packet.h"
#include "data.h"
#include "network.h"
#include "priority.h"

/****************************************************************************/
/* main header part */
/****************************************************************************/
/*!\class Scheduler sched.h
 * \brief Scheduler-Class used by the Polling deamon. This class schedules the device function calls in order of their priorities. It also takes care of the data storage for the Server. This class also contains some functions which are called from within the device-function like
 *int Scheduler::store(int )
 * int Scheduler::store(double)
 * int Scheduler::store(int *)
 * int Scheduler::store(double *)
 * int Scheduler::maximal_channels      (void )
 * int Scheduler::maximal_integrations  (void )
 * and others ..
 */
class Scheduler  : public QWidget {

  Q_OBJECT        //this Macro is for the signal/slot mechanism of Qt

public :
  // This is the Constructor. It initializes to normal mode with external trigger(=0)
  Scheduler(int = 100, int = 5, int = 1, int = 1);
  //This is the decstructor.
  ~Scheduler();

  // these two are called by the devices ....
  int error ( char *,char *formatstr, ... ); //new error funtion with format string
  /*! \fn int Scheduler::error ( char *formatstr, ... )
   * This function is called by a device routine to send and individual
   * error message to the error window for the user to react.
   * If the error widget appears, the user can cancel the measurement
   * (return value 0) or press ignore (return value : 1)
   * This function is called if the programmer wants
   * to print some formatted data in the error widget.
   * To get some infrmation on the format-commands, see the C-function printf.
   * \param char *formatstr, ... A  format string as used in the printf command
   * \return 0 The user has pressed the "abort" button.
   * \return 1 the user has pressed the "ignore" button
   */
  void init();
  void remove();
  void remove_continue();
  /*! This function is called before the actual measurement starts.
   *It calls the INIT_Routines of all devices sto prepare teh measurement
   */
  int  init_measurement (List *, Pr_list &);
  /*! This function is called after the measurement is finished.
   *It calls the CLOSE_Routines of all devices
   */
  int  close_measurement(List *, Pr_list &);
  //these are the routines that return after one cycle
  /*! This routine performs the measurement and calls the device functions*/
  int  measure(List *, Pr_list &);      //return 0 to force immedeate stop
  /*!This routine is used for the  measurement triggered by the internal trigger.
   * In case of external trigger ::measure(List *,Pr_list &) is called
   */
  int  SLOT_measure(List *, Pr_list &); //measurement routine called by timer event
  /*! Do some initialization for measurement based on internal trigger*/
  void initialize_SLOT       (void );   //reset parameters for timer controlled mesurement
  void set_max_channel_number(int  );   //set max_channels
  /*! \fn void Scheduler::set_max_channel_number(int  )
   * This function sets the maximal channel number. */
  void set_max_integrations  (int  );   //set the maximal number of inegrations
  //Set the current channel number.

  
  int  channel_number        (void );	//!Returns actual channel number.
  //!Return max_integrations.
  int  maximal_integrations  (void );
  //!Returns the maximal channel number.
  int  maximal_channels      (void );
  //! This function returns the actual channel
  int  act_channel           (void );  //returns the actual channel
  //! This functions returns the actual number of integrations.
  int  act_integrations      (void );  //return current number of integrations

  //!returns the trigger mode 0 = int, 1 = external
  int  trigger_mode          (void );
  //!returns monitor mode flag TRUE = on, FALSE = off
  int  monitor_mode          (void );

  /*!Store an integer value in the database for measured values.
  *  Call this function if you want to store measured integers
  */
  void store   ( int      );
  /*! Store a double value in the database for measured values.
  *   Call this function if you want to store measured double values.
  */
  void store   ( double   );
  /*! \brief Store an array of integer in the database for measured values.
  * Call this function if you want to store an array. (This can be the readout from a Digititzing Oscilloskope, for example). Check the file laser.cpp for an example on how to use this function.
  * \param  int array[], the array.
  * \param  int length, the length.
  * \return void.
  */
  void store   ( int *,int);
  /*!same as ::store( int *,int) only for arrays of double */
  void store( double *, int);
  void store   ( char *   );     //filename for image loading from disk
  void external_trigger      (void );  //set trigger external
  void internal_trigger      (void );  //set trigger internal
  void monitor_mode_on       (void );
  void monitor_mode_off      (void );
  void set_channel_number    (int  );
  void set_triggermode       (int  );   //trigger = 0 -> internal trigger, 1-> external
  void reset_integrations    (void );
  void reset_queue           (void );  //clear packet buffer list
  void reset                 (void );  //initialize for measurement
  uint delay                 (void );  //return delay
  void SetWaitFlag         (int,int); //wait for users error response
  void SetErrAnswer           (int );
  void set_cmd_sfd            (int );  //set Socket filedesriptor
  void call_device           (int  );  //call function with priority
  void max_channelnr         (int  );  //next functions are to access the according variables

  void inc_integrations      (void );   //increment actual integrations by 1
  void inc_channel           (void );   //increment actual channel number by one
  void reset_channel_number  (void );

  int DeQueue                 (int );  //send all buffer data to the client
  int  Number_devices();	             //Number of devices in Dev_Function Vector
  signals :
  void Sig_Error( char * );
  void channel_number_changed(int );
  void plot_data_point(uint ,int ); /*Send value of data point to plot window*/
  void plot_array(uint,int *);       //send array to plot routine
  void send_data(int, uint, int);    //send data to be stacked
  void send_data(uint, int, int*);   //send arrays,-> parameters :
  void send_data(QImage *,int);      //window_index,channel,size,pointer to array

public slots:
  void stop_measurement(int);
  void SLOT_setdelay(uint );
 //called when timeout is generated from internal trigger
  void SLOT_timeout( );

  //backup structures for integration of data
  //
  /*variable type 1 = int, 2 = double, 3 = char, 4 = char * ->*/
                  /*bad C++ code !!! I know -> It's your job to make it better*/
  private :

  struct backup {
    int    i_var;
    double d_var;
    int    type;
  };

struct arraybackup { //used  for integration of arrays
    int *iptr;
    int  size;
  };


  //List of measured data packets see : data.cpp
  //this class also contains functions to convert data to packets
  PacketList         *Que;
  dataPacket         dat_packet;  //Packet for sending measured data  -> see packet.h
  int max_channels;
  int max_integrations;
  int channels;                 //current channel
  int integrations;             //maximum number of integrations per cycle
  int priority;                 //actual priority currently called
  int stop ;
  int s_mode;                   //False if no signals for display are called from scheduler (high speed)
  int t_mode;                   //TRUE = external trigger ,FALSE = internal
  int device_status;            //return status of devices -> (used by SLOT_measure only)
  uint period;                  //delay for internal trigger contains ms
  bool busy;                    //busy flag to block parallel measurements
  char             *typelist;
  int               var_index;  //index off all "stored" variables
  int               temp_var_index;
  int               element_counter;   //contains the numberof measured data stored by device functions
  int               array_counter;     //counter for array-backup array
  int               temp_counter;      //temp counter for add_data function
  int               array_temp_counter;//temp counter for array add_data function
  backup           *buffPtr;
  arraybackup      *arrayPtr;    //array of pointers to int for integration of intarrays
  int *             backup_array;  //array for plotting-> divison by integrations
  //
  //end of database variables
  //
  //measurement control
  //
  bool            wait;         //Wait Flag for Scheduler->toggled by SLOT_Timeout
  int             data_saved;   //=0 if data is not yet saved
  int             monitor;      //monitor mode TRUE = on, FALSE = off
  int             maxarraysize; //size of largest array currently stored in database
  int             status;       //return status of measurement routines
  int             close_status; //return status of close routines
  int             err_wait;
  int             err_answer;
  int         	  cmd_sfd;     //socket fileds for sending error msg
};
 /*! \fn  int  Scheduler::maximal_integrations  (void)
  * returns the maximal number of integrations
  * \param no parameters
  * \return an integer
  */
  /*! \fn  int  Scheduler::maximal_channel  (void)
  * returns maximal channel number set by the user
  * \param no parameters
  * \return the channel number
  */
#endif
//don't write  behind the endif
